from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from tach.cache.setup import resolve_dot_tach

if TYPE_CHECKING:
    from pathlib import Path


def get_uid(project_root: Path) -> uuid.UUID | None:
    info_path = project_root / ".tach" / "tach.info"
    if not info_path.exists():
        resolve_dot_tach(project_root)
    contents = info_path.read_text().strip()
    uid = uuid.UUID(contents)
    return uid


def get_latest_version(project_root: Path) -> str | None:
    latest_version_path = project_root / ".tach" / ".latest-version"
    if not latest_version_path.exists():
        return
    version = latest_version_path.read_text().strip()
    return version



from __future__ import annotations

from typing import TYPE_CHECKING

from tach.filesystem.service import mark_executable, write_file
from tach.hooks import build_pre_commit_hook_content

if TYPE_CHECKING:
    from pathlib import Path


def install_pre_commit(project_root: Path) -> tuple[bool, str]:
    git_hooks_dir = project_root / ".git" / "hooks"
    hook_dst = git_hooks_dir / "pre-commit"

    if not git_hooks_dir.exists():
        return False, f"'{git_hooks_dir}' directory does not exist"

    if hook_dst.exists():
        return False, f"'{hook_dst}' already exists, you'll need to install manually"

    pre_commit_hook_content = build_pre_commit_hook_content()

    write_file(hook_dst, pre_commit_hook_content, root=project_root)
    mark_executable(hook_dst)
    return True, ""



from __future__ import annotations

import pytest

from tach.interactive.modules import FileNode, FileTree


@pytest.fixture
def project_root(tmp_path):
    (tmp_path / "dir1").mkdir()
    (tmp_path / "dir2").mkdir()
    (tmp_path / "dir1" / "__init__.py").touch()
    (tmp_path / "dir1" / "file1.py").touch()
    (tmp_path / "dir2" / "file2.py").touch()
    return tmp_path


def test_build_from_path(project_root):
    tree = FileTree.build_from_path(project_root)
    assert isinstance(tree, FileTree)
    assert tree.root.full_path == project_root
    assert tree.root.is_dir
    assert str(project_root / "dir1") in tree.nodes
    assert str(project_root / "dir2") in tree.nodes
    assert str(project_root / "dir1" / "file1.py") in tree.nodes
    assert str(project_root / "dir2" / "file2.py") in tree.nodes


def test_visible_children_property(project_root):
    node = FileNode.build_from_path(project_root / "dir1")
    node.expanded = True
    child_node = FileNode.build_from_path(project_root / "dir1" / "file1.py")
    node.children.append(child_node)
    assert node.visible_children == [child_node]
    node.expanded = False
    assert node.visible_children == []


def test_set_modules(project_root):
    tree = FileTree.build_from_path(project_root)
    tree.initialize_modules([project_root / "dir1" / "file1.py"], set())
    assert tree.nodes[str(project_root / "dir1" / "file1.py")].is_module


def test_set_source_root(project_root):
    tree = FileTree.build_from_path(project_root)
    new_source_root = tree.nodes[str(project_root / "dir2")]
    tree.initialize_source_roots([new_source_root.full_path])
    assert new_source_root.is_source_root


def test_siblings_method(project_root):
    tree = FileTree.build_from_path(project_root)
    node = tree.nodes[str(project_root / "dir1" / "file1.py")]
    siblings = node.siblings(include_self=True)
    assert node in siblings
    siblings = node.siblings(include_self=False)
    assert node not in siblings


def test_exclude_single_file(project_root):
    exclude_paths = ["dir1/file1.py"]
    tree = FileTree.build_from_path(project_root, exclude_paths=exclude_paths)
    assert str(project_root / "dir1" / "file1.py") not in tree.nodes
    assert str(project_root / "dir1") in tree.nodes


def test_exclude_entire_directory(project_root):
    exclude_paths = ["dir2/", "dir2/**"]
    tree = FileTree.build_from_path(project_root, exclude_paths=exclude_paths)
    assert str(project_root / "dir2") not in tree.nodes
    assert str(project_root / "dir1") in tree.nodes
    assert str(project_root / "dir2" / "file2.py") not in tree.nodes
    assert str(project_root / "dir2" / "nested_dir") not in tree.nodes


def test_exclude_nested_directory(project_root):
    exclude_paths = ["dir2/nested_dir/"]
    tree = FileTree.build_from_path(project_root, exclude_paths=exclude_paths)
    assert str(project_root / "dir2") in tree.nodes
    assert str(project_root / "dir2" / "nested_dir") not in tree.nodes
    assert str(project_root / "dir2" / "nested_dir" / "file4.py") not in tree.nodes


def test_exclude_multiple_patterns(project_root):
    exclude_paths = ["dir1/", "dir1/**", "dir2/nested_dir/"]
    tree = FileTree.build_from_path(project_root, exclude_paths=exclude_paths)
    assert str(project_root / "dir1") not in tree.nodes
    assert str(project_root / "dir2") in tree.nodes
    assert str(project_root / "dir2" / "nested_dir") not in tree.nodes



from enum import Enum
from pathlib import Path
from typing import Literal

class PythonImport:
    module_path: str
    line_number: int

def get_project_imports(
    project_root: Path,
    source_roots: list[Path],
    file_path: Path,
    project_config: ProjectConfig,
) -> list[PythonImport]: ...
def get_external_imports(
    project_root: Path,
    source_roots: list[Path],
    file_path: Path,
    project_config: ProjectConfig,
) -> list[PythonImport]: ...
def create_dependency_report(
    project_root: Path,
    project_config: ProjectConfig,
    path: Path,
    include_dependency_modules: list[str] | None,
    include_usage_modules: list[str] | None,
    skip_dependencies: bool,
    skip_usages: bool,
    raw: bool,
) -> str: ...
def create_computation_cache_key(
    project_root: Path,
    source_roots: list[Path],
    action: str,
    py_interpreter_version: str,
    file_dependencies: list[str],
    env_dependencies: list[str],
    backend: str,
    respect_gitignore: bool,
) -> str: ...
def check_computation_cache(
    project_root: Path, cache_key: str
) -> tuple[list[tuple[int, str]], int] | None: ...
def update_computation_cache(
    project_root: Path, cache_key: str, value: tuple[list[tuple[int, str]], int]
) -> None: ...
def parse_project_config(filepath: Path) -> tuple[ProjectConfig, bool]: ...
def parse_project_config_from_pyproject(filepath: Path) -> ProjectConfig: ...
def dump_project_config_to_toml(project_config: ProjectConfig) -> str: ...
def check(
    project_root: Path,
    project_config: ProjectConfig,
    dependencies: bool,
    interfaces: bool,
) -> list[Diagnostic]: ...
def check_external_dependencies(
    project_root: Path,
    project_config: ProjectConfig,
) -> list[Diagnostic]: ...
def format_diagnostics(
    project_root: Path,
    diagnostics: list[Diagnostic],
) -> str: ...
def detect_unused_dependencies(
    project_root: Path,
    project_config: ProjectConfig,
) -> list[UnusedDependencies]: ...
def sync_project(
    project_root: Path,
    project_config: ProjectConfig,
    add: bool = False,
) -> None: ...
def run_server(project_root: Path, project_config: ProjectConfig) -> None: ...
def serialize_modules_json(modules: list[ModuleConfig]) -> str: ...

class Diagnostic:
    def is_code(self) -> bool: ...
    def is_configuration(self) -> bool: ...
    def is_dependency_error(self) -> bool: ...
    def is_interface_error(self) -> bool: ...
    def is_warning(self) -> bool: ...
    def is_error(self) -> bool: ...
    def is_deprecated(self) -> bool: ...
    def usage_module(self) -> str | None: ...
    def definition_module(self) -> str | None: ...
    def to_string(self) -> str: ...
    def pyfile_path(self) -> str | None: ...
    def pyline_number(self) -> int | None: ...

def serialize_diagnostics_json(
    diagnostics: list[Diagnostic], pretty_print: bool
) -> str: ...

class DependencyConfig:
    path: str
    deprecated: bool

class ModuleConfig:
    path: str
    depends_on: list[DependencyConfig] | None
    cannot_depend_on: list[DependencyConfig] | None
    depends_on_external: list[str] | None
    cannot_depend_on_external: list[str] | None
    visibility: list[str]
    strict: bool
    unchecked: bool
    layer: str | None

    def __new__(cls, path: str, strict: bool) -> ModuleConfig: ...
    def mod_path(self) -> str: ...

InterfaceDataTypes = Literal["all", "primitive"]

class InterfaceConfig:
    expose: list[str]
    # 'from' in tach.toml
    from_modules: list[str]
    visibility: list[str] | None
    data_types: InterfaceDataTypes

CacheBackend = Literal["disk"]

class CacheConfig:
    backend: CacheBackend
    file_dependencies: list[str]
    env_dependencies: list[str]

class ExternalDependencyConfig:
    exclude: list[str]
    rename: list[str]

class UnusedDependencies:
    path: str
    dependencies: list[DependencyConfig]

RuleSetting = Literal["error", "warn", "off"]

RootModuleTreatment = Literal["allow", "ignore", "dependenciesonly", "forbid"]

class RulesConfig:
    unused_ignore_directives: RuleSetting
    require_ignore_directive_reasons: RuleSetting

class ProjectConfig:
    cache: CacheConfig
    external: ExternalDependencyConfig
    exclude: list[str]
    source_roots: list[str]
    exact: bool
    disable_logging: bool
    ignore_type_checking_imports: bool
    include_string_imports: bool
    forbid_circular_dependencies: bool
    respect_gitignore: bool
    # [DEPRECATED] Just reading this to print a warning
    use_regex_matching: bool
    rules: RulesConfig
    root_module: RootModuleTreatment

    def __new__(cls) -> ProjectConfig: ...
    def serialize_json(self) -> str: ...
    def exists(self) -> bool: ...
    def set_location(self, location: Path) -> None: ...
    def has_no_modules(self) -> bool: ...
    def has_no_dependencies(self) -> bool: ...
    def has_root_module_reference(self) -> bool: ...
    def module_paths(self) -> list[str]: ...
    def utility_paths(self) -> list[str]: ...
    def create_module(self, path: str) -> None: ...
    def delete_module(self, path: str) -> None: ...
    def mark_module_as_utility(self, path: str) -> None: ...
    def unmark_module_as_utility(self, path: str) -> None: ...
    def add_dependency(self, path: str, dependency: str) -> None: ...
    def remove_dependency(self, path: str, dependency: str) -> None: ...
    def add_source_root(self, path: Path) -> None: ...
    def remove_source_root(self, path: Path) -> None: ...
    def save_edits(self) -> None: ...
    def all_modules(self) -> list[ModuleConfig]: ...
    def all_interfaces(self) -> list[InterfaceConfig]: ...
    def filtered_modules(self, included_paths: list[Path]) -> list[ModuleConfig]: ...

class TachPytestPluginHandler:
    removed_test_paths: set[str]
    all_affected_modules: set[str]
    num_removed_items: int
    tests_ran_to_completion: bool
    def __new__(
        cls,
        project_root: Path,
        project_config: ProjectConfig,
        changed_files: list[Path],
        all_affected_modules: set[Path],
    ) -> TachPytestPluginHandler: ...
    def remove_test_path(self, path: Path) -> None: ...
    def should_remove_items(self, file_path: Path) -> bool: ...

class Direction(Enum):
    Dependencies = 0
    Dependents = 1

class DependentMap:
    def __new__(
        cls,
        project_root: Path,
        project_config: ProjectConfig,
        direction: Direction | None = None,
    ) -> DependentMap: ...
    def rebuild(self) -> None: ...
    def get_closure(self, paths: list[Path]) -> set[str]: ...
    def update_files(self, changed_files: list[Path]) -> None: ...
    def write_to_file(self, path: Path) -> None: ...
    def write_to_stdout(self) -> None: ...



from __future__ import annotations

from pathlib import Path

from tach.constants import CONFIG_FILE_NAME


def build_project_config_path(root: Path, file_name: str = CONFIG_FILE_NAME) -> Path:
    return root / f"{file_name}.toml"


def get_project_config_path(
    root: Path, *, file_name: str = CONFIG_FILE_NAME
) -> Path | None:
    file_path = build_project_config_path(root, file_name)
    if file_path.exists():
        return file_path
    return None


def get_pyproject_config_path(root: Path) -> Path | None:
    file_path = root / "pyproject.toml"
    if file_path.exists() and "tool.tach" in file_path.read_text():
        return file_path
    return None


def get_deprecated_project_config_path(root: Path | None = None) -> Path | None:
    root = root or Path.cwd()
    file_path = root / f"{CONFIG_FILE_NAME}.yaml"
    if file_path.exists():
        return file_path
    file_path = root / f"{CONFIG_FILE_NAME}.yml"
    if file_path.exists():
        return file_path
    return None


def has_project_config(root: Path) -> bool:
    return (
        get_project_config_path(root) is not None
        or get_pyproject_config_path(root) is not None
    )


def find_project_config_root() -> Path | None:
    cwd = Path.cwd()

    if has_project_config(cwd):
        return cwd

    # Iterate upwards, looking for project config
    for parent in cwd.parents:
        if has_project_config(parent):
            return parent

    return None



from __future__ import annotations

from tach.parsing.config import (
    combine_exclude_paths,
    dump_project_config_to_toml,
    parse_project_config,
)

__all__ = [
    "parse_project_config",
    "dump_project_config_to_toml",
    "combine_exclude_paths",
]



from __future__ import annotations

import fnmatch
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator
    from pathlib import Path


def _with_optional_trailing_slashes(patterns: list[str]) -> Generator[str, None, None]:
    for pattern in patterns:
        yield pattern
        if pattern.endswith("/"):
            yield pattern[:-1]


# Assumes 'relative_path' is a path relative to the project root
def is_path_excluded(exclude_paths: list[str], relative_path: Path) -> bool:
    if not exclude_paths:
        return False

    return any(
        (fnmatch.fnmatch(str(relative_path), exclude_path))
        for exclude_path in _with_optional_trailing_slashes(exclude_paths)
    )



from __future__ import annotations

import json
import shutil
from unittest.mock import NonCallableMagicMock

import pytest

from tach import icons
from tach.cli import tach_check, tach_check_external
from tach.errors import TachCircularDependencyError, TachVisibilityError
from tach.extension import Diagnostic
from tach.icons import FAIL, SUCCESS, WARNING
from tach.parsing.config import parse_project_config


def test_valid_example_dir(example_dir, capfd):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None
    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 0
    captured = capfd.readouterr()
    assert SUCCESS in captured.err
    assert WARNING in captured.err or "WARN" in captured.err


def test_valid_example_dir_monorepo(example_dir):
    project_root = example_dir / "monorepo_namespace"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None
    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 0


def test_check_json_output(example_dir, capfd, mocker):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    mock_diagnostics = [NonCallableMagicMock(spec=Diagnostic)]
    mock_diagnostics[0].is_error.return_value = False
    mocker.patch(
        "tach.extension.serialize_diagnostics_json",
        return_value=json.dumps([{"hello": "world"}]),
    )
    mocker.patch("tach.extension.check", return_value=mock_diagnostics)

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
            output_format="json",
        )
    assert exc_info.value.code == 0

    captured = capfd.readouterr()
    assert json.loads(captured.out) == [{"hello": "world"}]


def test_check_json_with_errors(example_dir, capfd, mocker):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    mock_diagnostics = [NonCallableMagicMock(spec=Diagnostic)]
    mocker.patch(
        "tach.extension.serialize_diagnostics_json",
        return_value=json.dumps(
            {"errors": ["error1", "error2"], "warnings": ["warning1"]}
        ),
    )
    mocker.patch("tach.extension.check", return_value=mock_diagnostics)

    with pytest.raises(SystemExit):
        tach_check(
            project_root=project_root,
            project_config=project_config,
            output_format="json",
        )

    captured = capfd.readouterr()
    assert json.loads(captured.out) == {
        "errors": ["error1", "error2"],
        "warnings": ["warning1"],
    }


def test_check_circular_dependency_text(example_dir, capfd, mocker):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    mocker.patch(
        "tach.extension.check",
        side_effect=TachCircularDependencyError(["mod1", "mod2", "mod1"]),
    )

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    assert "Circular dependency detected" in captured.err
    assert "'mod1'" in captured.err
    assert "'mod2'" in captured.err


def test_check_circular_dependency_json(example_dir, capfd, mocker):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    mocker.patch(
        "tach.extension.check",
        side_effect=TachCircularDependencyError(["mod1", "mod2", "mod1"]),
    )

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
            output_format="json",
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    result = json.loads(captured.out)
    assert result["error"] == "Circular dependency"
    assert result["dependencies"] == ["mod1", "mod2", "mod1"]


def test_check_visibility_error_text(example_dir, capfd, mocker):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    visibility_errors = [("mod1", "mod2", ["public"])]
    mocker.patch(
        "tach.extension.check", side_effect=TachVisibilityError(visibility_errors)
    )

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    assert "Module configuration error" in captured.err
    assert "'mod1' cannot depend on 'mod2'" in captured.err
    assert "public" in captured.err


def test_check_visibility_error_json(example_dir, capfd, mocker):
    project_root = example_dir / "valid"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    visibility_errors = [("mod1", "mod2", ["public"])]
    mocker.patch(
        "tach.extension.check", side_effect=TachVisibilityError(visibility_errors)
    )

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
            output_format="json",
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    result = json.loads(captured.out)
    assert result["error"] == "Visibility error"
    assert result["visibility_errors"] == [["mod1", "mod2", ["public"]]]


def test_distributed_config_example_dir(example_dir, capfd):
    project_root = example_dir / "distributed_config"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    assert FAIL in captured.err or "FAIL" in captured.err
    assert "Cannot use 'project.module_one.module_one'" in captured.err
    assert (
        "Module 'project.top_level' cannot depend on 'project.module_one'"
        in captured.err
    )
    assert "project/top_level.py" in captured.err


def _check_expected_messages_unordered(
    section_text: str, expected_messages: list[tuple]
) -> None:
    """Helper to verify all expected messages appear in a section of output text.

    Args:
        section_text: The text section to check
        expected_messages: List of tuples containing substrings that should appear together in a line
    """
    lines = iter(section_text.split("\n"))
    substr_tuples = set(expected_messages)
    for line in lines:
        if (
            "[WARN]" in line
            or "[FAIL]" in line
            or icons.FAIL in line
            or icons.WARNING in line
        ):
            matched = False
            for substr_tuple in substr_tuples:
                if all(substr.lower() in line.lower() for substr in substr_tuple):
                    substr_tuples.remove(substr_tuple)
                    matched = True
                    break
            if not matched:
                assert False, f"Unexpected warning/error line: {line}"

    assert not substr_tuples, (
        f"Not all expected messages were found: {list(substr_tuples)} in section: {section_text}"
    )


def test_many_features_example_dir(example_dir, capfd):
    project_root = example_dir / "many_features"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    general_header = captured.err.index("General\n")
    interfaces_header = captured.err.index("Interfaces\n")
    dependencies_header = captured.err.index("Internal Dependencies\n")
    unused_header = captured.err.index("Unused Dependencies")

    general_section = captured.err[general_header:interfaces_header]
    interfaces_section = captured.err[interfaces_header:dependencies_header]
    dependencies_section = captured.err[dependencies_header:unused_header]
    unused_section = captured.err[unused_header:]

    expected_general = [
        (
            "[WARN]",
            "other_src_root/module1/api.py",
            "ignore directive",
            "missing a reason",
        ),
        (
            "[WARN]",
            "other_src_root/module4/service.py",
            "ignore directive",
            "missing a reason",
        ),
        ("[WARN]", "real_src/main.py", "ignore directive", "missing a reason"),
        (
            "[FAIL]",
            "other_src_root/module4/service.py",
            "L6",
            "ignore directive",
            "unused",
        ),
        (
            "[FAIL]",
            "other_src_root/module4/service.py",
            "L12",
            "ignore directive",
            "unused",
        ),
        ("[FAIL]", "other_src_root/module1/api.py", "ignore directive", "unused"),
        ("[FAIL]", "real_src/main.py", "ignore directive", "unused"),
    ]

    expected_interfaces = [
        (
            "[FAIL]",
            "real_src/module1/__init__.py",
            "module3.anything",
            "public interface",
        ),
        (
            "[FAIL]",
            "real_src/module1/controller.py",
            "module5.something",
            "public interface",
        ),
        (
            "[FAIL]",
            "real_src/module1/controller.py",
            "module3.anything",
            "public interface",
        ),
        (
            "[FAIL]",
            "other_src_root/module5/__init__.py",
            "module1.api.something",
            "public interface",
        ),
    ]

    expected_dependencies = [
        ("[FAIL]", "real_src/module2/service.py", "outer_module", "module2"),
        ("[FAIL]", "real_src/module3/__init__.py", "'low'", "lower than", "'mid'"),
        (
            "[FAIL]",
            "real_src/globbed/other/module.py",
            "'hightest'",
            "lower than",
            "'high'",
        ),
        (
            "[FAIL]",
            "other_src_root/module4/service.py",
            "cannot depend on",
            "globbed.other.module.something",
        ),
        (
            "[FAIL]",
            "real_src/module1/controller.py",
            "'hightest'",
            "lower than",
            "'mid'",
            "module5",
        ),
        (
            "[FAIL]",
            "real_src/module1/controller.py",
            "L5",
            "'hightest'",
            "lower than",
            "'low'",
            "module3",
        ),
        (
            "[FAIL]",
            "real_src/module1/controller.py",
            "L6",
            "'hightest'",
            "lower than",
            "'low'",
            "module3",
        ),
        (
            "[FAIL]",
            "real_src/module3/submodule1/__init__.py",
            "'module3.submodule1'",
            "cannot depend on",
            "'module3.submodule2'",
        ),
    ]

    expected_unused = [
        ("module1", "module5"),
    ]

    _check_expected_messages_unordered(general_section, expected_general)
    _check_expected_messages_unordered(interfaces_section, expected_interfaces)
    _check_expected_messages_unordered(dependencies_section, expected_dependencies)
    _check_expected_messages_unordered(unused_section, expected_unused)


def test_many_features_example_dir__external(example_dir, capfd):
    project_root = example_dir / "many_features"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check_external(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    general_header = captured.err.index("General\n")
    external_header = captured.err.index("External Dependencies\n")

    general_section = captured.err[general_header:external_header]
    external_section = captured.err[external_header:]

    expected_general = [
        ("[WARN]", "real_src/main.py", "ignore directive", "missing a reason"),
        (
            "[WARN]",
            "real_src/module1/__init__.py",
            "ignore directive",
            "missing a reason",
        ),
        ("[FAIL]", "real_src/module1/__init__.py", "ignore directive", "unused"),
    ]

    expected_external = [
        ("[FAIL]", "prompt_toolkit", "not used"),
        ("[FAIL]", "importlib_metadata", "not used"),
        ("[FAIL]", "tomli_w", "not used"),
        ("[FAIL]", "pydot", "not used"),
        ("[FAIL]", "rich", "not used"),
        ("[FAIL]", "stdlib_list", "not used"),
        ("[FAIL]", "other_src_root/module5/__init__.py", "module5", "networkx"),
        ("[FAIL]", "real_src/module1/__init__.py", "module1", "tomli"),
        ("[FAIL]", "real_src/django_settings.py", "django", "not declared"),
        ("[FAIL]", "real_src/module3/content.py", "django", "not declared"),
    ]

    _check_expected_messages_unordered(general_section, expected_general)
    _check_expected_messages_unordered(external_section, expected_external)


def test_monorepo_workspace_example_dir(example_dir, capfd):
    project_root = example_dir / "monorepo_workspace"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 0

    captured = capfd.readouterr()
    configuration_header = captured.err.index("Configuration\n")

    configuration_section = captured.err[configuration_header:]

    expected_configuration = [
        ("[WARN]", "No first-party imports"),
    ]

    _check_expected_messages_unordered(configuration_section, expected_configuration)


def test_monorepo_workspace_example_dir_external(example_dir, capfd):
    project_root = example_dir / "monorepo_workspace"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check_external(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    external_header = captured.err.index("External Dependencies\n")

    external_section = captured.err[external_header:]

    expected_external = [
        ("[FAIL]", "requests", "not used"),
        ("[FAIL]", "src/albatross/__init__.py", "package2", "not declared"),
    ]

    _check_expected_messages_unordered(external_section, expected_external)


def test_visibility_error_example_dir(example_dir, capfd):
    project_root = example_dir / "visibility_error"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    assert "Module configuration error" in captured.err
    assert "'module2' cannot depend on 'module3'" in captured.err
    assert "module3" in captured.err
    assert "['module1']" in captured.err


def test_many_features_example_dir_with_gitignore(example_dir, capfd, tmp_path):
    project_root = tmp_path / "many_features"
    shutil.copytree(example_dir / "many_features", project_root)

    (project_root / ".gitignore").write_text("""
# Ignore module3 entirely
real_src/module3/
# Ignore specific files
real_src/module1/controller.py
other_src_root/module4/service.py
""")

    (project_root / "real_src" / ".gitignore").write_text("""
# Ignore all python files in globbed directory
globbed/**/*.py
""")

    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    general_header = captured.err.index("General\n")
    interfaces_header = captured.err.index("Interfaces\n")
    dependencies_header = captured.err.index("Internal Dependencies\n")
    unused_header = captured.err.index("Unused Dependencies")

    general_section = captured.err[general_header:interfaces_header]
    interfaces_section = captured.err[interfaces_header:dependencies_header]
    dependencies_section = captured.err[dependencies_header:unused_header]
    unused_section = captured.err[unused_header:]

    # Files that are gitignored should not appear in the diagnostics
    assert "real_src/module3/" not in captured.err
    assert "real_src/module1/controller.py" not in captured.err
    assert "other_src_root/module4/service.py" not in captured.err
    assert "real_src/globbed/" not in captured.err

    expected_general = [
        (
            "[WARN]",
            "other_src_root/module1/api.py",
            "ignore directive",
            "missing a reason",
        ),
        ("[WARN]", "real_src/main.py", "ignore directive", "missing a reason"),
        ("[FAIL]", "other_src_root/module1/api.py", "ignore directive", "unused"),
        ("[FAIL]", "real_src/main.py", "ignore directive", "unused"),
    ]

    expected_interfaces = [
        (
            "[FAIL]",
            "other_src_root/module5/__init__.py",
            "module1.api.something",
            "public interface",
        ),
    ]

    expected_dependencies = [
        ("[FAIL]", "real_src/module2/service.py", "outer_module", "module2"),
    ]

    expected_unused = [
        ("module1", "module5"),
    ]

    _check_expected_messages_unordered(general_section, expected_general)
    _check_expected_messages_unordered(interfaces_section, expected_interfaces)
    _check_expected_messages_unordered(dependencies_section, expected_dependencies)
    _check_expected_messages_unordered(unused_section, expected_unused)


def test_many_features_example_dir_with_gitignore__external(
    example_dir, capfd, tmp_path
):
    project_root = tmp_path / "many_features"
    shutil.copytree(example_dir / "many_features", project_root)

    (project_root / ".gitignore").write_text("""
# Ignore module3 entirely
real_src/module3/
# Ignore specific files
real_src/module1/controller.py
other_src_root/module4/service.py
""")

    (project_root / "real_src" / ".gitignore").write_text("""
# Ignore all python files in globbed directory
globbed/**/*.py
""")

    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    with pytest.raises(SystemExit) as exc_info:
        tach_check_external(
            project_root=project_root,
            project_config=project_config,
        )
    assert exc_info.value.code == 1

    captured = capfd.readouterr()
    general_header = captured.err.index("General\n")
    external_header = captured.err.index("External Dependencies\n")

    general_section = captured.err[general_header:external_header]
    external_section = captured.err[external_header:]

    # Files that are gitignored should not appear in the diagnostics
    assert "real_src/module3/" not in captured.err
    assert "real_src/module1/controller.py" not in captured.err
    assert "other_src_root/module4/service.py" not in captured.err
    assert "real_src/globbed/" not in captured.err

    expected_general = [
        ("[WARN]", "real_src/main.py", "ignore directive", "missing a reason"),
        (
            "[WARN]",
            "real_src/module1/__init__.py",
            "ignore directive",
            "missing a reason",
        ),
        ("[FAIL]", "real_src/module1/__init__.py", "ignore directive", "unused"),
    ]

    expected_external = [
        ("[FAIL]", "prompt_toolkit", "not used"),
        ("[FAIL]", "importlib_metadata", "not used"),
        ("[FAIL]", "tomli_w", "not used"),
        ("[FAIL]", "pydot", "not used"),
        ("[FAIL]", "rich", "not used"),
        ("[FAIL]", "stdlib_list", "not used"),
        ("[FAIL]", "other_src_root/module5/__init__.py", "module5", "networkx"),
        ("[FAIL]", "real_src/module1/__init__.py", "module1", "tomli"),
        ("[FAIL]", "real_src/django_settings.py", "django", "not declared"),
    ]

    _check_expected_messages_unordered(general_section, expected_general)
    _check_expected_messages_unordered(external_section, expected_external)



from __future__ import annotations

import json
import logging
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, Dict

from tach import __version__
from tach.cache import get_uid
from tach.logging.worker import create_managed_subprocess

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger("tach")
logger.setLevel(logging.INFO)


def init_logging(project_root: Path) -> None:
    remote_handler = RemoteLoggingHandler(project_root)
    logger.addHandler(remote_handler)


@dataclass
class CallInfo:
    function: str
    parameters: Dict[str, Any] = field(default_factory=dict)


class RemoteLoggingHandler(logging.Handler):
    def __init__(self, project_root: Path):
        super().__init__()
        self.uid = get_uid(project_root)
        self.file_path = create_managed_subprocess(project_root)

    def emit(self, record: logging.LogRecord) -> None:
        log_entry = self.format(record)
        with open(self.file_path, "a") as f:
            json.dump(
                {
                    "version": __version__,
                    "uid": str(self.uid) if self.uid else None,
                    "call_info": asdict(getattr(record, "data"))
                    if hasattr(record, "data")
                    else {},
                    "level": record.levelname,
                    "timestamp": record.created,
                    "log_entry": log_entry,
                },
                f,
            )
            f.write("\n")



from __future__ import annotations

import json

from tach.modularity import export_report
from tach.parsing.config import parse_project_config


def test_export_report_diagnostics(example_dir, tmp_path):
    example_project = example_dir / "many_features"

    project_config = parse_project_config(root=example_project)
    assert project_config is not None

    output_path = tmp_path / "modularity_report.json"

    export_report(example_project, project_config, output_path, force=True)

    assert output_path.exists()

    with open(output_path) as f:
        report = json.load(f)

    assert "diagnostics" in report
    assert isinstance(report["diagnostics"], str)

    diagnostics = json.loads(report["diagnostics"])

    assert isinstance(diagnostics, list)
    assert len(diagnostics) > 0

    for diagnostic in diagnostics:
        assert len(diagnostic) == 1
        variant = next(iter(diagnostic))
        assert variant in ("Located", "Global")

        if variant == "Located":
            located = diagnostic[variant]
            assert "file_path" in located
            assert "line_number" in located
            assert "severity" in located
            assert "details" in located
            assert located["severity"] in ("Error", "Warning")

            details = located["details"]
            assert len(details) == 1
            detail_type = next(iter(details))
            assert detail_type in ("Code", "Configuration")

        elif variant == "Global":
            global_diag = diagnostic[variant]
            assert "severity" in global_diag
            assert "details" in global_diag
            assert global_diag["severity"] in ("Error", "Warning")

            details = global_diag["details"]
            assert len(details) == 1
            detail_type = next(iter(details))
            assert detail_type in ("Code", "Configuration")



from __future__ import annotations

import shutil
import tempfile
from pathlib import Path

import pytest

from tach.cli import tach_sync
from tach.parsing.config import parse_project_config


def test_valid_example_dir(example_dir, capfd):
    project_root = example_dir / "valid"

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_project_root = Path(temp_dir) / "valid"
        shutil.copytree(project_root, temp_project_root)

        project_config = parse_project_config(root=temp_project_root)
        assert project_config is not None

        with pytest.raises(SystemExit) as exc_info:
            tach_sync(
                project_root=temp_project_root,
                project_config=project_config,
                add=True,
            )

        assert exc_info.value.code == 0
        captured = capfd.readouterr()
        assert "✅" in captured.err  # success state


def test_distributed_config_dir(example_dir, capfd):
    project_root = example_dir / "distributed_config"

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_project_root = Path(temp_dir) / "distributed_config"
        shutil.copytree(project_root, temp_project_root)

        project_config = parse_project_config(root=temp_project_root)
        assert project_config is not None

        modules = project_config.all_modules()
        assert len(modules) == 3

        top_level_module = next(
            module for module in modules if module.path == "project.top_level"
        )
        assert set(map(lambda dep: dep.path, top_level_module.depends_on)) == {
            "project.module_two"
        }

        with pytest.raises(SystemExit) as exc_info:
            tach_sync(
                project_root=temp_project_root,
                project_config=project_config,
            )

        assert exc_info.value.code == 0
        captured = capfd.readouterr()
        assert "✅" in captured.err  # success state

        project_config = parse_project_config(root=temp_project_root)
        assert project_config is not None

        modules = project_config.all_modules()
        assert len(modules) == 3

        top_level_module = next(
            module for module in modules if module.path == "project.top_level"
        )
        assert set(map(lambda dep: dep.path, top_level_module.depends_on)) == {
            "project.module_two",
            "project.module_one",
        }


def test_many_features_example_dir(example_dir, capfd):
    project_root = example_dir / "many_features"

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_project_root = Path(temp_dir) / "many_features"
        shutil.copytree(project_root, temp_project_root)

        project_config = parse_project_config(root=temp_project_root)
        assert project_config is not None

        with pytest.raises(SystemExit) as exc_info:
            tach_sync(
                project_root=temp_project_root,
                project_config=project_config,
            )

        assert exc_info.value.code == 0
        captured = capfd.readouterr()
        assert "✅" in captured.err  # success state

        project_config = parse_project_config(root=temp_project_root)
        assert project_config is not None

        modules = project_config.all_modules()
        # This should be the number of statically defined modules, before globbing
        assert len(modules) == 15

        module2 = next(module for module in modules if module.path == "module2")
        assert set(map(lambda dep: dep.path, module2.depends_on)) == {"outer_module"}

        module3 = next(module for module in modules if module.path == "module3")
        assert set(map(lambda dep: dep.path, module3.depends_on)) == {"module1"}

        assert (
            '"//module1"'
            in (
                temp_project_root / "real_src" / "module3" / "tach.domain.toml"
            ).read_text()
        )

        assert (
            '"<domain_root>"'
            in (
                temp_project_root / "real_src" / "module3" / "tach.domain.toml"
            ).read_text()
        )



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from enum import Enum
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Generator

from prompt_toolkit import ANSI
from prompt_toolkit.application import Application
from prompt_toolkit.data_structures import Point
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.layout import (
    Container,
    HSplit,
    Layout,
    ScrollablePane,
    VerticalAlign,
    Window,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import Frame
from rich.console import Console
from rich.text import Text
from rich.tree import Tree

from tach import errors
from tach import filesystem as fs
from tach.utils.exclude import is_path_excluded

if TYPE_CHECKING:
    from prompt_toolkit.formatted_text import AnyFormattedText

    from tach.extension import ProjectConfig


@dataclass
class FileNode:
    full_path: Path
    is_dir: bool
    expanded: bool = False
    is_module: bool = False
    is_source_root: bool = False
    is_utility: bool = False
    parent: FileNode | None = None
    children: list[FileNode] = field(default_factory=list)

    @property
    def empty(self) -> bool:
        return len(self.children) == 0

    @property
    def visible_children(self) -> list[FileNode]:
        if not self.expanded:
            return []
        return self.children

    @classmethod
    def build_from_path(cls, path: Path) -> FileNode:
        is_dir = path.is_dir()
        return cls(full_path=path, is_dir=is_dir)

    @property
    def parent_sorted_children(self) -> list[FileNode] | None:
        if not self.parent:
            return None
        return sorted(self.parent.visible_children, key=lambda node: node.full_path)

    @property
    def prev_sibling(self) -> FileNode | None:
        parent_sorted_children = self.parent_sorted_children
        if not parent_sorted_children:
            return None

        try:
            my_index = parent_sorted_children.index(self)
        except ValueError:
            raise errors.TachError("Error occurred in interactive file tree navigation")

        if my_index == 0:
            return None
        return parent_sorted_children[my_index - 1]

    @property
    def next_sibling(self) -> FileNode | None:
        parent_sorted_children = self.parent_sorted_children
        if not parent_sorted_children:
            return None

        try:
            my_index = parent_sorted_children.index(self)
        except ValueError:
            raise errors.TachError("Error occurred in interactive file tree navigation")

        if my_index == len(parent_sorted_children) - 1:
            return None
        return parent_sorted_children[my_index + 1]

    def siblings(self, include_self: bool = True) -> list[FileNode]:
        if not self.parent:
            return [self] if include_self else []

        return (
            self.parent.children
            if include_self
            else [node for node in self.parent.children if node is not self]
        )

    def clean_flags(self):
        self.is_module = False
        self.is_source_root = False
        self.is_utility = False


@dataclass
class FileTree:
    root: FileNode
    nodes: dict[str, FileNode] = field(default_factory=dict)

    @classmethod
    def build_from_path(
        cls,
        path: Path,
        depth: int | None = 1,
        exclude_paths: list[str] | None = None,
    ) -> FileTree:
        root = FileNode.build_from_path(path)
        root.is_module = False
        root.expanded = True
        tree = cls(root=root)
        tree.nodes[str(path)] = root
        tree._build_subtree(
            root,
            depth=depth if depth is not None else 1,
            exclude_paths=exclude_paths,
        )
        return tree

    def _build_subtree(
        self,
        root: FileNode,
        depth: int = 1,
        exclude_paths: list[str] | None = None,
    ):
        if root.is_dir:
            try:
                for entry in root.full_path.iterdir():
                    if entry.name.startswith("."):
                        # Ignore hidden files and directories
                        continue
                    if entry.is_file() and not entry.name.endswith(".py"):
                        # Only interested in Python files
                        continue

                    if entry.name == "__init__.py":
                        # __init__.py does not have a unique module path from its containing package
                        # so users should not be able to mark it as a standalone module
                        continue

                    # Exclude patterns are relative to project root, and may include a trailing slash

                    if exclude_paths and is_path_excluded(
                        exclude_paths, entry.relative_to(self.root.full_path)
                    ):
                        # This path is ignored
                        continue
                    child_node = FileNode.build_from_path(entry)
                    if depth > 1:
                        child_node.expanded = True
                    child_node.parent = root
                    root.children.append(child_node)
                    self.nodes[str(entry)] = child_node
                    if child_node.is_dir:
                        self._build_subtree(
                            child_node,
                            depth=max(depth - 1, 0),
                            exclude_paths=exclude_paths,
                        )
            except PermissionError:
                # This is expected to occur during iterdir when the directory cannot be accessed
                # We simply bail if that happens, meaning it won't show up in the interactive viewer
                return

    def expand_all_parent_dirs(self, node: FileNode) -> None:
        curr_node = node
        while curr_node.parent is not None and curr_node.parent.is_dir:
            curr_node.parent.expanded = True
            curr_node = curr_node.parent

    def initialize_modules(self, module_paths: list[Path], utility_paths: set[Path]):
        # NOTE: module_paths here are filesystem paths; they may be files or dirs
        for module_path in module_paths:
            module_path_str = str(module_path)
            if module_path_str in self.nodes:
                node = self.nodes[module_path_str]
                if module_path in utility_paths:
                    node.is_utility = True
                else:
                    node.is_module = True
                self.expand_all_parent_dirs(node)

    def initialize_source_roots(self, source_roots: list[Path]):
        # NOTE: assuming source_roots are absolute here
        for source_root in source_roots:
            if str(source_root) not in self.nodes:
                continue
            node = self.nodes[str(source_root)]
            node.clean_flags()
            node.is_source_root = True
            node.expanded = True
            self.expand_all_parent_dirs(node)

    def __iter__(self):
        return file_tree_iterator(self)

    def visible(self):
        return file_tree_iterator(self, visible_only=True)


def file_tree_iterator(
    tree: FileTree, visible_only: bool = False
) -> Generator[FileNode, None, None]:
    # DFS traversal for printing
    stack = deque([tree.root])

    while stack:
        node = stack.popleft()
        yield node
        if visible_only:
            stack.extendleft(
                sorted(node.visible_children, key=lambda n: n.full_path, reverse=True)
            )
        else:
            stack.extendleft(
                sorted(node.children, key=lambda n: n.full_path, reverse=True)
            )


class ExitCode(Enum):
    QUIT_NOSAVE = 1
    QUIT_SAVE = 2


@dataclass
class InteractiveModuleConfiguration:
    source_roots: list[Path]
    module_paths: list[Path]
    utility_paths: list[Path]


class InteractiveModuleTree:
    TREE_LABEL = "Confirm Your Modules"

    def __init__(
        self,
        path: Path,
        project_config: ProjectConfig,
        exclude_paths: list[str],
        depth: int | None = 1,
    ):
        # By default, don't save if we exit for any reason
        self.exit_code: ExitCode = ExitCode.QUIT_NOSAVE
        self.file_tree = FileTree.build_from_path(
            path=path, depth=depth, exclude_paths=exclude_paths
        )

        source_roots = [
            path / source_root for source_root in project_config.source_roots
        ]
        module_file_paths = list(
            filter(
                None,
                [
                    fs.module_to_pyfile_or_dir_path(
                        source_roots=tuple(source_roots),
                        module_path=module_path,
                    )
                    for module_path in project_config.module_paths()
                ],
            ),
        )
        utility_file_paths = set(
            filter(
                None,
                [
                    fs.module_to_pyfile_or_dir_path(
                        source_roots=tuple(source_roots),
                        module_path=module_path,
                    )
                    for module_path in project_config.utility_paths()
                ],
            ),
        )
        self.file_tree.initialize_modules(
            module_paths=module_file_paths, utility_paths=utility_file_paths
        )
        self.selected_node = self.file_tree.root

        self.file_tree.initialize_source_roots(source_roots=source_roots)

        # x location doesn't matter, only need to track hidden cursor for auto-scroll behavior
        # y location starts at 1 because the FileTree is rendered with a labeled header above the first branch
        self.cursor_point = Point(x=0, y=1)
        self.console = Console()
        self.tree_control = FormattedTextControl(
            text=ANSI(self._render_tree()),
            focusable=True,
            show_cursor=False,
            get_cursor_position=self.get_cursor_position_fn(),
        )
        self.footer_control = self._build_footer()
        self.layout = Layout(
            HSplit(
                [
                    Frame(ScrollablePane(Window(self.tree_control))),
                    self.footer_control,
                ]
            )
        )
        self.key_bindings = KeyBindings()
        self._register_keybindings()
        self.styles = self._build_styles()
        self.app: Application[None] = Application(
            layout=self.layout,
            key_bindings=self.key_bindings,
            full_screen=True,
            style=self.styles,
        )

    def get_cursor_position_fn(self) -> Callable[[], Point]:
        def get_cursor_position() -> Point:
            return self.cursor_point

        return get_cursor_position

    def move_cursor_up(self):
        self.cursor_point = Point(x=self.cursor_point.x, y=self.cursor_point.y - 1)

    def move_cursor_down(self):
        self.cursor_point = Point(x=self.cursor_point.x, y=self.cursor_point.y + 1)

    @staticmethod
    def _build_styles() -> Style:
        return Style.from_dict(
            {
                "footer-key": "bold cyan",
            }
        )

    KEY_BINDING_LEGEND_TOP: list[tuple[str, str]] = [
        ("Up/Down", "Navigate"),
        ("Right", "Expand"),
        ("Left", "Collapse"),
        ("Ctrl + Up", "Jump to parent"),
    ]
    KEY_BINDING_LEGEND_MIDDLE: list[tuple[str, str]] = [
        ("Enter", "Mark/unmark Module"),
        ("s", "Mark/unmark Source Root"),
        ("u", "Mark/unmark Utility"),
    ]
    KEY_BINDING_LEGEND_BOTTOM: list[tuple[str, str]] = [
        ("Ctrl + s", "Exit and save"),
        ("Ctrl + c", "Exit without saving"),
        ("Ctrl + a", "Mark/unmark all"),
    ]

    @staticmethod
    def _key_binding_text(binding: str, description: str) -> list[tuple[str, str]]:
        return [("class:footer-key", binding), ("", f": {description}  ")]

    @classmethod
    def _build_footer(cls) -> Container:
        def _build_footer_text(
            bindings: list[tuple[str, str]],
        ) -> AnyFormattedText:
            return list(
                chain(
                    *(
                        cls._key_binding_text(binding[0], binding[1])
                        for binding in bindings
                    )
                )
            )

        footer_text_top: AnyFormattedText = _build_footer_text(
            cls.KEY_BINDING_LEGEND_TOP
        )
        footer_text_middle = _build_footer_text(cls.KEY_BINDING_LEGEND_MIDDLE)
        footer_text_bottom: AnyFormattedText = _build_footer_text(
            cls.KEY_BINDING_LEGEND_BOTTOM
        )
        return HSplit(
            [
                Window(
                    FormattedTextControl(text=footer_text_top), dont_extend_height=True
                ),
                Window(
                    FormattedTextControl(text=footer_text_middle),
                    dont_extend_height=True,
                ),
                Window(
                    FormattedTextControl(text=footer_text_bottom),
                    dont_extend_height=True,
                ),
            ],
            align=VerticalAlign.CENTER,
        )

    def _register_keybindings(self):
        if self.key_bindings.bindings:
            return

        @self.key_bindings.add("c-c")
        def _(event: KeyPressEvent):
            self.exit_code = ExitCode.QUIT_NOSAVE
            self.app.exit()

        @self.key_bindings.add("c-s")
        def _(event: KeyPressEvent):
            self.exit_code = ExitCode.QUIT_SAVE
            self.app.exit()

        @self.key_bindings.add("up")
        def _(event: KeyPressEvent):
            prev_sibling = self.selected_node.prev_sibling
            # If previous sibling exists, want to bubble down to last child of this sibling
            if prev_sibling:
                curr_node = prev_sibling
                while curr_node.visible_children:
                    curr_node = sorted(
                        curr_node.visible_children, key=lambda node: node.full_path
                    )[-1]
                self.selected_node = curr_node
                self.move_cursor_up()
                self._update_display()
            # If no previous sibling, go to parent
            elif self.selected_node.parent:
                self.selected_node = self.selected_node.parent
                self.move_cursor_up()
                self._update_display()

        @self.key_bindings.add("down")
        def _(event: KeyPressEvent):
            # If we have children, should go to first child alphabetically
            if self.selected_node.visible_children:
                self.selected_node = sorted(
                    self.selected_node.visible_children, key=lambda node: node.full_path
                )[0]
                self.move_cursor_down()
                self._update_display()
                return
            # If we have no children and no parent, nothing to do
            elif not self.selected_node.parent:
                return

            # Here we need to bubble up to find the next node
            curr_node = self.selected_node
            next_sibling = self.selected_node.next_sibling
            while next_sibling is None:
                if not curr_node.parent:
                    break
                curr_node = curr_node.parent
                next_sibling = curr_node.next_sibling

            if not next_sibling:
                # We are the last child all the way up to root
                return

            self.selected_node = next_sibling
            self.move_cursor_down()
            self._update_display()

        @self.key_bindings.add("right")
        def _(event: KeyPressEvent):
            self.selected_node.expanded = True
            self._update_display()

        @self.key_bindings.add("left")
        def _(event: KeyPressEvent):
            self.selected_node.expanded = False
            self._update_display()

        @self.key_bindings.add("enter")
        def _(event: KeyPressEvent):
            if self.selected_node is self.file_tree.root:
                # Root should not be explicitly selected
                return

            if not self.selected_node.is_module:
                # Flags are exclusive
                self.selected_node.clean_flags()
                self.selected_node.is_module = True
            else:
                self.selected_node.is_module = False
            self._update_display()

        @self.key_bindings.add("s")
        def _(event: KeyPressEvent):
            if not self.selected_node.is_source_root:
                # Flags are exclusive
                self.selected_node.clean_flags()
                self.selected_node.is_source_root = True
            else:
                self.selected_node.is_source_root = False
            self._update_display()

        @self.key_bindings.add("u")
        def _(event: KeyPressEvent):
            if not self.selected_node.is_utility:
                # Flags are exclusive
                self.selected_node.clean_flags()
                self.selected_node.is_utility = True
            else:
                self.selected_node.is_utility = False
            self._update_display()

        @self.key_bindings.add("c-a")
        def _(event: KeyPressEvent):
            if not self.selected_node.parent:
                # This means we are the root node without siblings
                # We should simply toggle ourselves
                self.selected_node.is_module = not self.selected_node.is_module
                self._update_display()
                return

            # If all siblings are currently modules, we should un-set all of them (target value is False)
            # Otherwise, we want to ensure all of them are set as modules (target value is True)
            all_siblings_are_modules = all(
                node.is_module for node in self.selected_node.siblings()
            )
            for node in self.selected_node.siblings():
                node.is_module = not all_siblings_are_modules

            self._update_display()

        @self.key_bindings.add("c-up")
        def _(event: KeyPressEvent):
            if not self.selected_node.parent:
                return

            # Simple way to keep cursor position accurate while jumping to parent
            while self.selected_node.prev_sibling:
                self.move_cursor_up()
                self.selected_node = self.selected_node.prev_sibling
            self.move_cursor_up()
            self.selected_node = self.selected_node.parent
            self._update_display()

    def _render_node(self, node: FileNode) -> Text:
        text_parts: list[tuple[str, str] | str] = []
        if node == self.selected_node:
            text_parts.append(("-> ", "bold cyan"))

        basename = node.full_path.name
        if node.is_source_root:
            text_parts.append((f"[Source Root] {basename}", "bold cyan"))
        elif node.is_module:
            text_parts.append((f"[Module] {basename}", "bold yellow"))
        elif node.is_utility:
            text_parts.append((f"[Utility] {basename}", "bold magenta"))
        elif node == self.selected_node:
            text_parts.append((basename, "bold"))
        else:
            text_parts.append(basename)

        if not node.empty and node.expanded:
            text_parts.append((" ∨", "cyan"))
        elif not node.empty:
            text_parts.append((" >", "cyan"))
        return Text.assemble(*text_parts)

    def _render_tree(self):
        tree_root = Tree(self.TREE_LABEL)
        # Mapping FileNode paths to rich.Tree branches
        # so that we can iterate over the FileTree and use the
        # parent pointers to find the parent rich.Tree branches
        tree_mapping: dict[str, Tree] = {}

        for node in self.file_tree.visible():
            if node.parent is None:
                # If no parent on FileNode, add to rich.Tree root
                tree_node = tree_root.add(self._render_node(node))
            else:
                if str(node.parent.full_path) not in tree_mapping:
                    raise errors.TachError("Failed to render module tree.")
                # Find parent rich.Tree branch,
                # attach this FileNode to the parent's branch
                parent_tree_node = tree_mapping[str(node.parent.full_path)]
                tree_node = parent_tree_node.add(self._render_node(node))

            # Add this new FileNode to the mapping
            tree_mapping[str(node.full_path)] = tree_node

        with self.console.capture() as capture:
            self.console.print(tree_root)
        return capture.get()

    def _update_display(self):
        self.tree_control.text = ANSI(self._render_tree())

    def run(self) -> InteractiveModuleConfiguration | None:
        self.app.run()
        if self.exit_code == ExitCode.QUIT_SAVE:
            module_paths: list[Path] = []
            source_roots: list[Path] = []
            utility_paths: list[Path] = []
            for node in self.file_tree:
                if node.is_module:
                    module_paths.append(node.full_path)
                elif node.is_source_root:
                    source_roots.append(node.full_path)
                elif node.is_utility:
                    utility_paths.append(node.full_path)

            return InteractiveModuleConfiguration(
                source_roots=source_roots,
                module_paths=module_paths,
                utility_paths=utility_paths,
            )


def get_selected_modules_interactive(
    path: Path,
    project_config: ProjectConfig,
    exclude_paths: list[str],
    depth: int | None = 1,
) -> InteractiveModuleConfiguration | None:
    ipt = InteractiveModuleTree(
        path=path,
        project_config=project_config,
        exclude_paths=exclude_paths,
        depth=depth,
    )
    return ipt.run()



from __future__ import annotations


class TachError(Exception): ...


class TachSetupError(TachError): ...


class TachConfigError(TachError): ...


class TachClosedBetaError(TachError): ...


class TachCircularDependencyError(TachError):
    def __init__(self, dependencies: list[str]):
        self.dependencies = dependencies
        super().__init__("Circular dependency error")


class TachVisibilityError(TachError):
    def __init__(self, visibility_errors: list[tuple[str, str, list[str]]]):
        self.visibility_errors = visibility_errors
        super().__init__("Visibility error")


class TachInitCancelledError(TachError): ...



from __future__ import annotations

import os
import sys
from enum import Enum
from functools import lru_cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


def is_interactive() -> bool:
    return sys.stdout.isatty() and sys.stderr.isatty()


class BCOLORS:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


def colorize(text: str, color_start: str, color_end: str = BCOLORS.ENDC) -> str:
    if is_interactive():
        return f"{color_start}{text}{color_end}"
    return text


class TerminalEnvironment(Enum):
    UNKNOWN = 1
    JETBRAINS = 2
    VSCODE = 3


@lru_cache(maxsize=None)
def detect_environment() -> TerminalEnvironment:
    if "jetbrains" in os.environ.get("TERMINAL_EMULATOR", "").lower():
        return TerminalEnvironment.JETBRAINS
    elif "vscode" in os.environ.get("TERM_PROGRAM", "").lower():
        return TerminalEnvironment.VSCODE
    return TerminalEnvironment.UNKNOWN


def render_path_simple(file_path: Path, line: int | None = None) -> str:
    if line is not None:
        return f"{file_path}[L{line}]"
    return str(file_path)


def create_clickable_link(
    file_path: Path, display_path: Path | None = None, line: int | None = None
) -> str:
    if not is_interactive():
        return render_path_simple(file_path, line)

    terminal_env = detect_environment()
    abs_path = file_path.resolve()

    if terminal_env == TerminalEnvironment.JETBRAINS:
        link = f"file://{abs_path}:{line}" if line is not None else f"file://{abs_path}"
    elif terminal_env == TerminalEnvironment.VSCODE:
        link = (
            f"vscode://file/{abs_path}:{line}"
            if line is not None
            else f"vscode://file/{abs_path}"
        )
    else:
        # For generic terminals, use a standard file link
        link = f"file://{abs_path}"

    # ANSI escape codes for clickable link
    if line:
        # Show the line number if we have it
        display_file_path = f"{display_path or file_path}[L{line}]"
    else:
        display_file_path = str(display_path) if display_path else str(file_path)
    clickable_link = f"\033]8;;{link}\033\\{display_file_path}\033]8;;\033\\"
    return clickable_link



from __future__ import annotations

import os

PACKAGE_NAME: str = "tach"
TOOL_NAME: str = "tach"
CONFIG_FILE_NAME: str = TOOL_NAME
PACKAGE_FILE_NAME: str = "package"
ROOT_MODULE_SENTINEL_TAG: str = "<root>"
DEFAULT_EXCLUDE_PATHS = [
    "**/tests",
    "**/docs",
    "**/*__pycache__",
    "**/*egg-info",
    "**/venv",
]

GAUGE_API_BASE_URL: str = os.getenv("GAUGE_API_BASE_URL", "https://app.gauge.sh")

__all__ = [
    "PACKAGE_NAME",
    "TOOL_NAME",
    "CONFIG_FILE_NAME",
    "PACKAGE_FILE_NAME",
    "ROOT_MODULE_SENTINEL_TAG",
    "DEFAULT_EXCLUDE_PATHS",
    "GAUGE_API_BASE_URL",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import tomli
import tomli_w

from tach import extension
from tach import filesystem as fs
from tach.constants import CONFIG_FILE_NAME

if TYPE_CHECKING:
    from pathlib import Path

    from tach.extension import ProjectConfig


def dump_project_config_to_toml(config: ProjectConfig) -> str:
    data = tomli.loads(extension.dump_project_config_to_toml(config))
    return tomli_w.dumps(data)


def migrate_deprecated_cache_backend(data: dict[str, Any]) -> dict[str, Any]:
    if "cache" in data:
        if "backend" in data["cache"]:
            data["cache"]["backend"] = "disk"
    return data


def migrate_deprecated_depends_on(data: dict[str, Any]) -> dict[str, Any]:
    if "modules" in data:
        for module in data["modules"]:
            if "depends_on" in module:
                for index, path in enumerate(module["depends_on"]):
                    if isinstance(path, str):
                        module["depends_on"][index] = {"path": path}
    return data


def migrate_deprecated_source_root(data: dict[str, Any]) -> dict[str, Any]:
    if "source_root" in data:
        if isinstance(data["source_root"], str):
            data["source_roots"] = [data["source_root"]]
            del data["source_root"]
    return data


def migrate_deprecated_yaml_config(filepath: Path) -> ProjectConfig:
    import yaml

    content = filepath.read_text()
    data = yaml.safe_load(content)

    try:
        data = migrate_deprecated_cache_backend(data)
        data = migrate_deprecated_depends_on(data)
        data = migrate_deprecated_source_root(data)
        toml_config = tomli_w.dumps(data)
        print("Auto-migrating deprecated YAML config to TOML...")
        filepath.with_suffix(".toml").write_text(toml_config)
        project_config, ext_migrated = extension.parse_project_config(
            filepath.with_suffix(".toml")
        )
        if ext_migrated:
            # This is a second migration pass, so we need to save the result
            filepath.with_suffix(".toml").write_text(
                dump_project_config_to_toml(project_config)
            )
    except TypeError as e:
        raise ValueError(f"Failed to parse deprecated YAML config: {e}")
    except ValueError as e:
        filepath.with_suffix(".toml").unlink()
        raise ValueError(f"Failed to parse deprecated YAML config: {e}")
    print("Deleting deprecated YAML config...")
    filepath.unlink()
    return project_config


def parse_project_config(
    root: Path,
    *,
    file_name: str = CONFIG_FILE_NAME,
) -> ProjectConfig | None:
    file_path = fs.get_project_config_path(root, file_name=file_name)
    if file_path:
        # Standard TOML config found
        project_config, ext_migrated = extension.parse_project_config(file_path)
        if ext_migrated:
            # Write the auto-migrated TOML config
            file_path.with_suffix(".toml").write_text(
                dump_project_config_to_toml(project_config)
            )
        return project_config
    elif (root / "pyproject.toml").exists():
        try:
            return extension.parse_project_config_from_pyproject(
                root / "pyproject.toml"
            )
        except Exception:
            return None
    else:
        # No TOML found, check for deprecated (YAML) config as a fallback
        file_path = fs.get_deprecated_project_config_path(root)
        if not file_path:
            return None
        # This will write the auto-migrated TOML config
        return migrate_deprecated_yaml_config(file_path)


def combine_exclude_paths(
    exclude_paths: list[str] | None,
    project_excludes: list[str],
) -> list[str]:
    if exclude_paths is not None:
        return list(set(exclude_paths + project_excludes))
    else:
        return project_excludes
